/*********************************************************************************************************************************
 * Copyright (c) Razvan Surdulescu 1996
 *               Erik Kunze        1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Razvan Surdulescu
 *          Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: slt.c,v 1.1 2000/07/13 18:34:06 erik Rel $";
#endif
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "util.h"
#include "z80.h"
#include "slt.h"
typedef struct _sltInfo {
	uns16 type;
	uns16 identifier;
	uns32 length;
} sltInfo;
typedef struct _sltData {
	off_t offsHeader, offsData;
	sltInfo *info;
} sltData;
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_SNAP) != 0) { x } }
#else
#define DEB(x)
#endif
static sltData sltDatas = { 0, 0, 0 };
int
SltRead(char *fname)
{
	int readonly;
	size_t size;
	void *mmp = (void *)-1;
	uns8 *scanBegin, *scanEnd, *scanPtr;
	sltData *sdp = &sltDatas;
	int blockcnt, blocknum, ret = -1;
	if (Z80Read(fname) < 0)
	{
		Msg(M_ERR, "SLT file does not contain a valid Z80 snapshot");
		goto quit;
	}
	readonly = 1;
	mmp = Mmap(fname, 1, &readonly, &size);
	if (mmp == (void *)-1)
	{
		goto quit;
	}
	scanBegin = (uns8 *)mmp;
	scanEnd = (uns8 *)mmp + size;
	for (scanPtr = scanBegin; scanPtr < scanEnd; scanPtr++)
	{
		if (memcmp(scanPtr, "\0\0\0SLT", 6) == 0)
		{
			break;
		}
	}
	if (scanPtr >= scanEnd)
	{
		Msg(M_ERR, "couldn't find SLT signature");
		goto quit;
	}
	scanPtr += 6;
	sdp->offsHeader = scanPtr - scanBegin;
	for (blockcnt = 1;; blockcnt++)
	{
		if (scanPtr + sizeof(sltInfo) > scanEnd)
		{
			Msg(M_ERR, "premature end on SLT snapshot");
			goto quit;
		}
		if (*scanPtr == 0x00 && *(scanPtr + 1) == 0x00)
		{
			break;
		}
		scanPtr += sizeof(sltInfo);
	}
	if (sdp->info != (sltInfo *)0)
	{
		free(sdp->info);
	}
	sdp->info = (sltInfo *)Malloc((size_t)blockcnt * sizeof(sltInfo), "SltRead");
	for (blocknum = 0, scanPtr = scanBegin + sdp->offsHeader; blocknum < blockcnt; blocknum++, scanPtr += sizeof(sltInfo))
	{
		sdp->info[blocknum].type = scanPtr[0] + (scanPtr[1] << 8);
		sdp->info[blocknum].identifier = scanPtr[2] + (scanPtr[3] << 8);
		sdp->info[blocknum].length = scanPtr[4] + (scanPtr[5] << 8) + (scanPtr[6] << 16) + (scanPtr[7] << 24);
		DEB(Msg(M_DEBUG, "block %2d, type %1d, id %2d, length %5d",
				blocknum, sdp->info[blocknum].type, sdp->info[blocknum].identifier, sdp->info[blocknum].length););
	}
	sdp->offsData = scanPtr - scanBegin;
	ret = 0;
  quit:
	if (mmp != (void *)-1)
	{
		Munmap(mmp, size);
	}
	return ret;
}
int
SltReadLevel(char *fname)
{
	int readonly;
	size_t size;
	void *mmp = (void *)-1;
	uns8 *scanBegin, *scanEnd, *scanPtr;
	sltData *sdp = &sltDatas;
	off_t disp;
	unsigned int which, addr, remain, count;
	int ret = -1;
	assert(sdp->info != (sltInfo *)0);
	for (which = 0, disp = 0; sdp->info[which].type != 0; which++)
	{
		if (sdp->info[which].identifier == (uns16)A)
		{
			break;
		}
		disp += (off_t)(sdp->info[which].length);
	}
	if (sdp->info[which].identifier != (uns16)A)
	{
		Msg(M_WARN, "requested level %d doesn't exist", A);
		goto quit;
	}
	readonly = 1;
	mmp = Mmap(fname, 1, &readonly, &size);
	if (mmp == (void *)-1)
	{
		goto quit;
	}
	scanBegin = (uns8 *)mmp;
	scanEnd = (uns8 *)mmp + size;
	addr = HL;
	remain = (unsigned int)(sdp->info[which].length);
	scanPtr = scanBegin + sdp->offsData + disp;
	if (scanPtr + remain > scanEnd)
	{
		Msg(M_PERR, "couldn't load level %d", A);
		goto quit;
	}
	while (scanPtr < scanEnd && remain > 0)
	{
		if (*scanPtr == 0xed && *(scanPtr + 1) == 0xed)
		{
			for (count = *(scanPtr + 2); count > 0 && addr < 65536; count--)
			{
				MEM(addr) = *(scanPtr + 3);
				addr++;
			}
			remain -= 4;
			scanPtr += 4;
		}
		else
		{
			MEM(addr) = *scanPtr;
			addr++;
			remain--;
			scanPtr++;
		}
	}
	if (remain > 0)
	{
		Msg(M_PWARN, "error while loading level %d", A);
		goto quit;
	}
	DEB(Msg(M_DEBUG, "reading level %2d of length %5d", sdp->info[which].identifier, sdp->info[which].length););
	ret = 0;
  quit:
	if (mmp != (void *)-1)
	{
		Munmap(mmp, size);
	}
	return ret;
}
int
SltWrite(char *fname, char *sltfname)
{
	FILE *ifp = (FILE *)0, *ofp = (FILE *)0;
	int byte, ret = -1;
	if (Z80Write(fname) < 0)
	{
		goto quit;
	}
	if ((ifp = Fopen(sltfname, "rb")) == (FILE *)0)
	{
		Msg(M_PERR, "couldn't open <%s> for reading", sltfname);
		goto quit;
	}
	if ((ofp = Fopen(fname, "ab")) == (FILE *)0)
	{
		Msg(M_PERR, "couldn't open <%s> for writing", fname);
		goto quit;
	}
	(void)putc(0, ofp);
	(void)putc(0, ofp);
	(void)putc(0, ofp);
	(void)putc('S', ofp);
	(void)putc('L', ofp);
	(void)putc('T', ofp);
	if (fseek(ifp, sltDatas.offsHeader, SEEK_SET) < 0)
	{
		goto write_error;
	}
	while ((byte = fgetc(ifp)) != EOF)
	{
		if (putc(byte, ofp) == EOF)
		{
		  write_error:
			Msg(M_PERR, "couldn't write SLT levels");
			goto quit;
		}
	}
	ret = 0;
  quit:
	if (ofp != (FILE *)0)
	{
		(void)fclose(ofp);
	}
	if (ifp != (FILE *)0)
	{
		(void)fclose(ifp);
	}
	return ret;
}
