/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: tzx.c,v 4.51 2000/12/18 16:39:19 erik Rel $";
#endif
#include <X11/keysym.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <assert.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "io.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#include "tape.h"
#include "tap.h"
#include "vtp.h"
#include "ui.h"
#include "tzx.h"
#define MAJREV							1
#define MINREV							13
#define MS								3500
#define INVALID							65535
#define NOW								(EmulVideoFrame * Machine->tstatesPerFrame + TSTATES)
enum { BS_PILOT, BS_SYNC1, BS_SYNC2, BS_BIT_1, BS_BIT_2, BS_PAUSE };
struct tzxHeader {
	char ident[7];
	char marker;
	unsigned char major;
	unsigned char minor;
};
#define LOW_PULSE						{ lastValue = 0; }
#define HIGH_PULSE						{ lastValue = B_EAR; }
#define TOGGLE_PULSE					{ lastValue ^= B_EAR; }
#define SET_IN_LOOP
#define RESET_IN_LOOP					{ loopStart = INVALID; }
#define IN_LOOP							( loopStart != INVALID )
#define SET_IN_CALL
#define RESET_IN_CALL					{ callStart = INVALID; }
#define IN_CALL							( callStart != INVALID )
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_TAPE) != 0) { x } }
#else
#define DEB(x)
#endif
static int tzxInNextBlock(void);
static unsigned int tzxInSelection(unsigned int, uns8 *);
static unsigned int blockType;
static int blockState;
static unsigned long lastTstate;
static uns16 loopStart, loopRepeat;
static uns16 callStart, callCount, callCurrent;
static uns16 pilotLen, pilot, sync1, sync2, bit0, bit1, curBit;
static uns32 blockLen;
static uns16 pauseLen;
static uns8 *blockPtr, lastByte, bitMask, bitsLeft, lastValue;
#define pulseLen						pilotLen
#define pulse							pilot
#define pulseCount						bitsLeft
#define sample							pilot
int
TzxInAttach(char *fname, uns8 **begin, uns8 **end)
{
	int readonly;
	size_t size;
	void *mmp;
	uns8 *scanBegin, *scanEnd, *scanPtr;
	struct tzxHeader *header;
	TzxRecord *tr;
	static const char *types[] = { "Program", "Num.Array", "Char.Array", "Bytes", "" };
	char buf[31];
	int block, len, type, i;
	int ret = -1;
	readonly = 1;
	mmp = Mmap(fname, 1, &readonly, &size);
	if (mmp == (void *)-1)
	{
		return -1;
	}
	header = (struct tzxHeader *)mmp;
	if (memcmp(header->ident, "ZXTape!", sizeof(header->ident)) != 0)
	{
		Msg(M_ERR, "input file is not in TZX format");
		goto quit;
	}
	DEB(Msg(M_DEBUG, "TZX: version %d.%d", header->major, header->minor););
	if (header->major > MAJREV || (header->major == MAJREV && header->minor > MINREV))
	{
		if (header->major == 10 && header->minor == 1)
		{
			Msg(M_WARN, "input file has invalid TZX version 10.1");
		}
		else
		{
			Msg(M_WARN, "XZX does not support TXZ format newer than %d.%d", MAJREV, MINREV);
			goto quit;
		}
	}
	RESET_IN_LOOP;
	RESET_IN_CALL;
	scanBegin = (uns8 *)mmp + sizeof(struct tzxHeader);
	scanEnd = (uns8 *)mmp + size;
	block = 0;
	for (scanPtr = scanBegin;;)
	{
		tr = Malloc(sizeof(TzxRecord), "TzxInAttach");
		TapeInSetRecord((TapeRecord *)tr);
		block++;
		tr->pos = scanPtr;
		if (scanPtr >= scanEnd)
		{
			tr->type = 0x00;
			tr->str = Strdup("***      End Of Tape      ***", "TzxInAttach");
			break;
		}
		tr->type = *scanPtr++;
		DEB(Msg(M_DEBUG, "TZX: scaning, id = %02x, pos = %6d", tr->type, tr->pos - scanBegin););
		switch (tr->type)
		{
			case 0x10:
				len = GET2(scanPtr, 0x02);
				scanPtr += 0x04;
				goto identify;
			case 0x11:
				len = GET3(scanPtr, 0x0f);
				scanPtr += 0x12;
			  identify:
				if (*scanPtr == 0 && len == 19)
				{
					scanPtr++;
					type = *scanPtr < 4 ? *scanPtr : 4;
					scanPtr++;
					for (i = 0; i < 10; i++)
					{
						buf[i] = (*scanPtr < ' ' || *scanPtr > 0x7f) ? '?' : *scanPtr;
						scanPtr++;
					}
					scanPtr += 7;
				}
				else
				{
					type = 4;
					memset(buf, ' ', 10);
					scanPtr += len;
				}
				(void)sprintf(&buf[10], "%c %-11s%6d", tr->type == 0x10 ? ' ' : '*', types[type], len >= 2 ? len - 2 : 0);
				tr->str = Strdup(buf, "TzxInAttach");
				break;
			case 0x12:
				tr->str = Strdup("Pure tone", "TzxInAttach");
				scanPtr += 0x04;
				break;
			case 0x13:
				tr->str = Strdup("Sequence of pulses", "TzxInAttach");
				scanPtr += 0x01 + (GET1(scanPtr, 0) * 2);
				break;
			case 0x14:
				len = GET3(scanPtr, 0x07);
				(void)sprintf(&buf[0], "Pure data%20d", len);
				tr->str = Strdup(buf, "TzxInAttach");
				scanPtr += 0x0a + len;
				break;
			case 0x15:
				tr->str = Strdup("Direct", "TzxInAttach");
				scanPtr += 0x08 + GET3(scanPtr, 0x05);
				break;
			case 0x20:
				tr->str = Strdup(GET2(scanPtr, 0x00) != 0 ? "Pause" : "Stop the tape", "TzxInAttach");
				scanPtr += 0x02;
				break;
			case 0x21:
				i = GET1(scanPtr, 0x00) > 23 ? 23 : GET1(scanPtr, 0x00);
				(void)sprintf(buf, "Group: %.*s", i, scanPtr + 0x01);
				tr->str = Strdup(buf, "TzxInAttach");
				scanPtr += 0x01 + GET1(scanPtr, 0);
				break;
			case 0x22:
				tr->str = Strdup("Group end", "TzxInAttach");
				scanPtr += 0x00;
				break;
			case 0x23:
				(void)sprintf(buf, "Jump %d", block + (sgn16)GET2(tr->pos, 1) - 1);
				tr->str = Strdup(buf, "TzxInAttach");
				scanPtr += 0x02;
				break;
			case 0x24:
				tr->str = Strdup("Loop start", "TzxInAttach");
				scanPtr += 0x02;
				break;
			case 0x25:
				tr->str = Strdup("Loop end", "TzxInAttach");
				scanPtr += 0x00;
				break;
			case 0x26:
				tr->str = Strdup("Call", "TzxInAttach");
				scanPtr += 0x02 + (GET2(scanPtr, 0x00) * 2);
				break;
			case 0x27:
				tr->str = Strdup("Return", "TzxInAttach");
				scanPtr += 0x00;
				break;
			case 0x28:
				tr->str = Strdup("Select", "TzxInAttach");
				scanPtr += 0x02 + GET2(scanPtr, 0x00);
				break;
			case 0x2a:
				tr->str = Strdup("Stop the tape if in 48k mode", "TzxInAttach");
				scanPtr += 0x04 + GET4(scanPtr, 0x00);
				break;
			case 0x30:
				i = GET1(scanPtr, 0x00);
				if (i > 30)
				{
					i = 30;
				}
				(void)sprintf(buf, "%.*s", i, scanPtr + 0x01);
				tr->str = Strdup(buf, "TzxInAttach");
				scanPtr += 0x01 + GET1(scanPtr, 0);
				break;
			case 0x31:
				tr->str = Strdup("Message", "TzxInAttach");
				scanPtr += 0x02 + GET1(scanPtr, 1);
				break;
			case 0x32:
				tr->str = Strdup("Archive info", "TzxInAttach");
				scanPtr += 0x02 + GET2(scanPtr, 0x00);
				break;
			case 0x33:
				tr->str = Strdup("Hardware", "TzxInAttach");
				scanPtr += 0x01 + *scanPtr * 3;
				break;
			case 0x34:
				tr->str = Strdup("Emulation info", "TzxInAttach");
				scanPtr += 0x08;
				break;
			case 0x35:
				(void)sprintf(buf, "%.16s", scanPtr);
				tr->str = Strdup(buf, "TzxInAttach");
#if 0
				if (strncmp(scanPtr, "POKEs           ", 16) == 0)
				{
				}
				else if (strncmp(scanPtr, "Instructions    ", 16) == 0)
				{
				}
				else if (strncmp(scanPtr, "Spectrum Screen ", 16) == 0)
				{
				}
#endif
				scanPtr += 0x14 + GET4(scanPtr, 0x10);
				break;
			case 0x40:
				tr->str = Strdup("Snapshot", "TzxInAttach");
				scanPtr += 0x04 + GET3(scanPtr, 0x01);
				break;
			case 0x5a:
				tr->str = Strdup("Merged files", "TzxInAttach");
				scanPtr += 0x09;
				break;
			default:
				Msg(M_WARN, "unknown TZX block %x", tr->type);
				tr->str = Strdup("Unknown", "TzxInAttach");
				scanPtr += 0x04 + (uns32)GET4(scanPtr, 0x00);
				break;
		}
	}
	*begin = scanBegin;
	*end = scanEnd;
	ret = 0;
  quit:
	if (ret != 0 && mmp != (void *)-1)
	{
		Munmap(mmp, size);
	}
	return ret;
}
void
TzxInPlay(void)
{
	TzxRecord *tr;
	tr = (TzxRecord *)TapeInGetRecord(TapeInGetBlock());
	assert(tr != (TzxRecord *)0);
	if (tr->type == 0x20 || tr->type == 0x2a)
	{
		TapeInNextBlock();
	}
	tzxInNextBlock();
	lastTstate = NOW;
	LOW_PULSE;
}
uns8
TzxInGetBit(void)
{
	unsigned int passed;
#ifdef REAL_AUDIO
	uns8 saveLastValue = lastValue;
#endif
	passed = NOW - lastTstate;
  again:
	switch (blockType)
	{
		case 0x10:
		case 0x11:
		case 0x14:
			if (blockState == BS_PILOT)
			{
				while (passed >= pilot)
				{
					passed -= pilot;
					lastTstate += pilot;
					TOGGLE_PULSE;
					if (--pilotLen == 0)
					{
						blockState = BS_SYNC1;
						goto sync1;
					}
				}
			}
			else if (blockState == BS_SYNC1)
			{
			  sync1:
				if (passed >= sync1)
				{
					passed -= sync1;
					lastTstate += sync1;
					TOGGLE_PULSE;
					blockState = BS_SYNC2;
					goto sync2;
				}
			}
			else if (blockState == BS_SYNC2)
			{
			  sync2:
				if (passed >= sync2)
				{
					passed -= sync2;
					lastTstate += sync2;
					TOGGLE_PULSE;
					if (blockLen != 0)
					{
						bitMask = 0x80;
						bitsLeft = blockLen == 1 ? lastByte : 8;
						curBit = (*blockPtr & bitMask) != 0 ? bit1 : bit0;
						blockState = BS_BIT_1;
						goto bit_1;
					}
					else
					{
						blockState = BS_PAUSE;
						goto pause;
					}
				}
			}
			else if (blockState == BS_BIT_1)
			{
			  bit_1:
				if (passed >= curBit)
				{
					passed -= curBit;
					lastTstate += curBit;
					TOGGLE_PULSE;
					blockState = BS_BIT_2;
					goto bit_2;
				}
			}
			else if (blockState == BS_BIT_2)
			{
			  bit_2:
				if (passed >= curBit)
				{
					passed -= curBit;
					lastTstate += curBit;
					TOGGLE_PULSE;
					if (--bitsLeft == 0)
					{
						if (--blockLen == 0)
						{
							if (pauseLen != 0)
							{
								blockState = BS_PAUSE;
								goto pause;
							}
							else
							{
								goto next;
							}
						}
						bitMask = 0x80;
						bitsLeft = blockLen == 1 ? lastByte : 8;
						blockPtr++;
					}
					else
					{
						bitMask >>= 1;
					}
					curBit = (*blockPtr & bitMask) != 0 ? bit1 : bit0;
					blockState = BS_BIT_1;
					goto bit_1;
				}
			}
			else if (blockState == BS_PAUSE)
			{
			  pause:
				while (passed >= 1*MS)
				{
					passed -= 1*MS;
					lastTstate += 1*MS;
					LOW_PULSE;
					if (--pauseLen == 0)
					{
					  next:
						TapeInNextBlock();
						if ((TapeGetState() & TS_AUTOSTOP) != 0)
						{
							TapeInStop();
						}
						else if (tzxInNextBlock() == 0)
						{
							goto again;
						}
					}
				}
			}
			break;
		case 0x12:
			while (passed >= pulse)
			{
				passed -= pulse;
				lastTstate += pulse;
				TOGGLE_PULSE;
				if (--pulseLen == 0)
				{
					goto next;
				}
			}
			break;
		case 0x13:
			while (passed >= pulse)
			{
				passed -= pulse;
				lastTstate += pulse;
				TOGGLE_PULSE;
				if (--pulseCount == 0)
				{
					goto next;
				}
				else
				{
					pulse = (uns16)GET2(blockPtr, 0x00);
					blockPtr += 2;
				}
			}
			break;
		case 0x15:
			while (passed >= sample)
			{
				passed -= sample;
				lastTstate += sample;
				if ((*blockPtr & bitMask) != 0)
				{
					HIGH_PULSE;
				}
				else
				{
					LOW_PULSE;
				}
				if (--bitsLeft == 0)
				{
					if (--blockLen == 0)
					{
						if (pauseLen != 0)
						{
							blockType = 0x10;
							blockState = BS_PAUSE;
							goto pause;
						}
						else
						{
							goto next;
						}
					}
					bitMask = 0x80;
					bitsLeft = blockLen == 1 ? lastByte : 8;
					blockPtr++;
				}
				else
				{
					bitMask >>= 1;
				}
			}
			break;
		case 0x20:
			goto pause;
	}
#ifdef REAL_AUDIO
	if (CFGGET(tapeNoise) != 0 && lastValue != saveLastValue)
	{
		SpeakerOutByte(lastValue >> 2);
	}
#endif
	return lastValue;
}
#ifndef NO_ED_TRAPS
int
TzxInReadBlock(void)
{
	while ((TapeGetState() & (TS_ERRIN | TS_ENDIN | TS_PLAY)) == TS_PLAY)
	{
		if (blockType > 0x00 && blockType < 0x20)
		{
			if (blockState < BS_BIT_1)
			{
				if (blockType != 0x10)
				{
					break;
				}
				if (blockLen == 0 || *blockPtr == A1)
				{
					TapInReadBlock(blockPtr, (int)blockLen);
					if (pauseLen != 0)
					{
						blockState = BS_PAUSE;
					}
					else
					{
						TapeInNextBlock();
						(void)tzxInNextBlock();
					}
					lastTstate = NOW;
					return 0;
				}
			}
			else if (blockState == BS_PAUSE)
			{
				blockState = BS_PILOT;
			}
			TapeInNextBlock();
		}
		(void)tzxInNextBlock();
	}
	return -1;
}
#endif
static int
tzxInNextBlock(void)
{
	TzxRecord *tr, *tr2;
	unsigned int block, blocks;
	int i;
	for (block = TapeInGetBlock(), blocks = TapeInGetBlocks(); block < blocks;)
	{
		VtpUpdate();
		tr = (TzxRecord *)TapeInGetRecord(block);
		assert(tr != (TzxRecord *)0);
		blockType = tr->type;
		DEB(Msg(M_DEBUG, "TZX: curBlock = %3u, blockType = %02x", block, blockType););
		switch (blockType)
		{
			case 0x00:
				DEB(Msg(M_DEBUG, "TZX: end of tape reached"););
				goto quit;
			case 0x10:
				pauseLen = (uns16)GET2(tr->pos, 0x01);
				blockLen = (uns16)GET2(tr->pos, 0x03);
				blockPtr = tr->pos + 0x05;
				pilotLen = *blockPtr != 0 ? 3220 : 8064;
				pilot = 2168;
				sync1 = 667;
				sync2 = 735;
				bit0 = 855;
				bit1 = 1710;
				lastByte = 8;
				blockState = BS_PILOT;
				goto quit;
			case 0x11:
				pauseLen = (uns16)GET2(tr->pos, 0x0e);
				blockLen = (uns32)GET3(tr->pos, 0x10);
				blockPtr = tr->pos + 0x13;
				pilotLen = (uns16)GET2(tr->pos, 0x0b);
				pilot = (uns16)GET2(tr->pos, 0x01);
				sync1 = (uns16)GET2(tr->pos, 0x03);
				sync2 = (uns16)GET2(tr->pos, 0x05);
				bit0 = (uns16)GET2(tr->pos, 0x07);
				bit1 = (uns16)GET2(tr->pos, 0x09);
				lastByte = GET1(tr->pos, 0x0d);
				blockState = BS_PILOT;
				goto quit;
			case 0x12:
				pulseLen = (uns16)GET2(tr->pos, 0x03);
				pulse = (uns16)GET2(tr->pos, 0x01);
				goto quit;
			case 0x13:
				pulseCount = GET1(tr->pos, 0x01);
				pulse = (uns16)GET2(tr->pos, 0x02);
				blockPtr = tr->pos + 0x04;
				goto quit;
			case 0x14:
				pauseLen = (uns16)GET2(tr->pos, 0x06);
				blockLen = (uns32)GET3(tr->pos, 0x08);
				blockPtr = tr->pos + 0x0b;
				bit0 = (uns16)GET2(tr->pos, 0x01);
				bit1 = (uns16)GET2(tr->pos, 0x03);
				lastByte = GET1(tr->pos, 0x05);
				bitMask = 0x80;
				bitsLeft = blockLen == 1 ? lastByte : 8;
				curBit = (*blockPtr & bitMask) != 0 ? bit1 : bit0;
				blockState = BS_BIT_1;
				goto quit;
			case 0x15:
				pauseLen = (uns16)GET2(tr->pos, 0x03);
				blockLen = (uns32)GET3(tr->pos, 0x06);
				blockPtr = tr->pos + 0x09;
				sample = (uns16)GET2(tr->pos, 0x01);
				sample = sample != 0 ? sample : 1;
				lastByte = GET1(tr->pos, 0x05);
				bitMask = 0x80;
				goto quit;
			case 0x20:
				pauseLen = (uns16)GET2(tr->pos, 0x01);
				if (pauseLen == 0)
				{
					TapeInStop();
				}
				goto quit;
			case 0x21:
			case 0x22:
				block++;
				break;
			case 0x23:
				block += (sgn16)GET2(tr->pos, 1);
				break;
			case 0x24:
				SET_IN_LOOP;
				loopStart = block;
				loopRepeat = (uns16)GET2(tr->pos, 1);
				block++;
				break;
			case 0x25:
				if (!IN_LOOP)
				{
					Msg(M_ERR, "Loop End without Loop in TZX file");
					TapeSetState(TapeGetState() | TS_ERRIN);
					TapeInStop();
					return -1;
				}
				if (--loopRepeat != 0)
				{
					block = loopStart + 1U;
				}
				else
				{
					block++;
					RESET_IN_LOOP;
				}
				break;
			case 0x26:
				SET_IN_CALL;
				callStart = block;
				callCount = (uns16)GET2(tr->pos, 1);
				callCurrent = 0;
				block += (sgn16)GET2(tr->pos, 3);
				break;
			case 0x27:
				if (!IN_CALL)
				{
					Msg(M_ERR, "Return without Call in TZX file");
					TapeSetState(TapeGetState() | TS_ERRIN);
					TapeInStop();
					return -1;
				}
				if (++callCurrent < callCount)
				{
					tr2 = (TzxRecord *)TapeInGetRecord((unsigned int)callStart);
					assert(tr2 != (TzxRecord *)0);
					i = 1 + 2 * callCurrent;
					block = (unsigned int)((sgn16)callStart + (sgn16)GET2(tr2->pos, i));
				}
				else
				{
					block = callStart + 1U;
					RESET_IN_CALL;
				}
				break;
			case 0x28:
				block = tzxInSelection(block, tr->pos);
				break;
			case 0x2a:
				if (CFGGET(machine) <= SP_48_3)
				{
					TapeInStop();
					goto quit;
				}
				block++;
				break;
			case 0x40:
				Msg(M_WARN, "unsupported TZX block %02x", blockType);
				block++;
				break;
			case 0x30:
			case 0x31:
			case 0x32:
			case 0x33:
			case 0x34:
			case 0x35:
			case 0x5a:
				block++;
				break;
			default:
				Msg(M_WARN, "unknown TZX block %02x", blockType);
				block++;
				break;
		}
		(void)TapeInSetBlock(block);
	}
  quit:
	if ((TapeGetState() & (TS_ENDIN | TS_ERRIN)) != 0)
	{
		TapeInStop();
		return -1;
	}
	return 0;
}
static unsigned int
tzxInSelection(unsigned int block, uns8 *bp)
{
	int cnt = GET1(bp, 0x03), i;
	int choice = 0;
	char buf[31];
	char **choices;
	char *s;
	if (cnt == 0)
	{
		Msg(M_WARN, "Selection with zero choices in TZX file");
		return ++block;
	}
	choices = Malloc((size_t)cnt * sizeof(char  *), "tzxInSelection");
	for (i = 0, s = (char *)&bp[0x04]; i < cnt; i++)
	{
		(void)sprintf(buf, "%.*s", s[2] > 30 ? 30 : (int)s[2], s + 3);
		choices[i] = Strdup(buf, "tzxInSelection");
		s += 3 + s[2];
	}
	if ((choice = UiBrowseList("Choose an item", choices, cnt, choice)) == -1)
	{
		choice = 0;
	}
	while (--cnt >= 0)
	{
		free(choices[cnt]);
	}
	free(choices);
	for (i = 0, s = (char *)&bp[0x04]; i < choice; i++)
	{
		s += 3 + s[2];
	}
	block += (sgn16)GET2(s, 0x00);
	return block;
}
