/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: tape.c,v 1.9 2000/11/26 18:52:53 erik Rel $";
#endif
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>
#include <assert.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#include "tap.h"
#include "tzx.h"
#include "tzxinfo.h"
#include "voc.h"
#include "vtp.h"
#include "tape.h"
typedef struct _tapeData {
	unsigned int state;
	struct {
		const char *fname;
		uns8 *begin, *end;
		Llist *recordList;
		unsigned int recordCount;
		unsigned int recordCurrent;
		EmulEventId event;
	} in;
	struct {
		const char *fname;
		FILE *fp;
	} out;
} tapeData;
#define TAPE_TIMER_VALUE				3
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_TAPE) != 0) { x } }
#else
#define DEB(x)
#endif
static void tapeDestroyRecord(void *);
static void tapeInGetBit(uns8 (*)(void));
static uns8 tapeInGetBitDummy(void);
static void tapeCatchFault(int);
uns8 (*TapeInGetBit)(void) = tapeInGetBitDummy;
static tapeData tapeDatas;
static jmp_buf tapeEnv;
void
TapeInit(void)
{
	tapeData *tdp = &tapeDatas;
	tdp->state = (CFGGET(tapeFastload) > 0 ? TS_FASTLOAD : 0) | (CFGGET(tapeAutostop) > 0 ? TS_AUTOSTOP : 0);
#ifndef NO_ED_TRAPS
	tdp->state |= CFGGET(tapeQuickload) > 0 ? TS_QUICKLOAD : 0;
#endif
#ifdef REAL_AUDIO
	tdp->state |= CFGGET(tapeNoise) > 0 ? TS_NOISE : 0;
#endif
	tdp->in.fname = (char *)0;
	tdp->in.recordList = (Llist *)0;
	tdp->in.recordCount = 0;
	tdp->in.event = (EmulEventId)0;
	tdp->out.fname = (char *)0;
}
unsigned int
TapeGetState(void)
{
	return tapeDatas.state;
}
void
TapeSetState(unsigned int newstate)
{
	tapeDatas.state = newstate;
}
unsigned int
TapeInGetBlock(void)
{
	return tapeDatas.in.recordCurrent;
}
int
TapeInSetBlock(unsigned int block)
{
	if (block >= tapeDatas.in.recordCount)
	{
		return -1;
	}
	tapeDatas.in.recordCurrent = block;
	if ((block + 1) == tapeDatas.in.recordCount)
	{
		tapeDatas.state |= TS_ENDIN;
	}
	else
	{
		tapeDatas.state &= ~TS_ENDIN;
	}
	return 0;
}
unsigned int
TapeInGetBlocks(void)
{
	return tapeDatas.in.recordCount;
}
int
TapeInNextBlock(void)
{
	return (TapeInSetBlock(tapeDatas.in.recordCurrent + 1));
}
TapeRecord *
TapeInGetRecord(unsigned int number)
{
	return ((TapeRecord *)RetrieveElemList(tapeDatas.in.recordList, number));
}
unsigned int
TapeInSetRecord(TapeRecord *record)
{
	tapeDatas.in.recordList = AppendElemList(tapeDatas.in.recordList, (void *)record);
	return ++tapeDatas.in.recordCount;
}
const char *
TapeInGetInfo(unsigned int number)
{
	assert((tapeDatas.state & (TS_TAPIN | TS_TZXIN | TS_VOCIN)) != 0);
	return (TzxGetInfo(number));
}
const char *
TapeInGetName(void)
{
	return tapeDatas.in.fname;
}
#ifndef NO_ED_TRAPS
void
TapeInReadBlock(void)
{
	tapeData *tdp = &tapeDatas;
	DEB(Msg(M_DEBUG, "%s request: id = %02x, addr = %04X, length = %5d", F1 & FLAG_C ? "load" : "verify", A1, IX, DE););
	if ((tdp->state & (TS_TAPIN | TS_TZXIN | TS_VOCIN)) == 0 || (tdp->state & TS_ENDIN) != 0)
	{
		VtpNewInputFile();
	}
	if ((tdp->state & (TS_TAPIN | TS_TZXIN | TS_VOCIN)) != 0 && (tdp->state & (TS_PLAY | TS_ENDIN | TS_ERRIN | TS_AUTOSTOP)) == 0)
	{
		TapeInPlay();
	}
	if 	((tdp->state & (TS_TAPIN | TS_TZXIN)) != 0 && (tdp->state  & (TS_PLAY | TS_QUICKLOAD)) == (TS_PLAY | TS_QUICKLOAD))
	{
		if (TzxInReadBlock() == 0)
		{
			return;
		}
	}
	TSTATES += 11 - 8;
	DEC_R();
	A = Z80_IN((A << 8) | 0xfe);
}
#endif
int
TapeInAttach(char *fname)
{
	struct sigaction newact, oldact;
	tapeData *tdp = &tapeDatas;
	unsigned int type = 0;
	uns8 *begin, *end;
	int ret = -1;
	TapeInDetach();
	newact.sa_handler = tapeCatchFault;
	(void)sigemptyset(&newact.sa_mask);
#ifdef SA_RESTART
	newact.sa_flags = SA_RESTART;
#else
	newact.sa_flags = 0;
#endif
	(void)sigaction(SIGSEGV, &newact, &oldact);
	if (setjmp(tapeEnv) == 0)
	{
		switch (GetFileType(fname))
		{
			case FT_TAP:
				type = TS_TAPIN;
				ret = TapInAttach(fname, &begin, &end);
				break;
			case FT_TZX:
				type = TS_TZXIN;
				ret = TzxInAttach(fname, &begin, &end);
				break;
			case FT_VOC:
				type = TS_VOCIN;
				ret = VocInAttach(fname, &begin, &end);
				break;
		}
		if (ret == 0)
		{
			tdp->state |= type | ((tdp->in.recordCurrent + 1) == tdp->in.recordCount ? TS_ENDIN : 0);
			tdp->in.fname = fname;
			tdp->in.begin = begin;
			tdp->in.end = end;
		}
	}
	else
	{
		Msg(M_ERR, "corrupted input file");
	}
	(void)sigaction(SIGSEGV, &oldact, (struct sigaction *)0);
	return ret;
}
static void
tapeDestroyRecord(void *entry)
{
	if (((TapeRecord *)entry)->str != (char *)0)
	{
		free(((TapeRecord *)entry)->str);
	}
}
void
TapeInDetach(void)
{
	tapeData *tdp = &tapeDatas;
	if ((tdp->state & (TS_TAPIN | TS_TZXIN | TS_VOCIN)) != 0)
	{
		assert((tdp->state & TS_PLAY) == 0);
		TapeInGetBit = tapeInGetBitDummy;
		tdp->state &= ~(TS_TAPIN | TS_TZXIN | TS_VOCIN | TS_PLAY | TS_ERRIN | TS_ENDIN);
		Munmap((void *)(tdp->in.begin), (size_t)(tdp->in.end - tdp->in.begin));
		free((char *)(tdp->in.fname));
		tdp->in.fname = (char *)0;
		DestroyList(tdp->in.recordList, tapeDestroyRecord);
		tdp->in.recordList = (Llist *)0;
		tdp->in.recordCount = 0;
		tdp->in.recordCurrent = 0;
	}
}
void
TapeInPlay(void)
{
	tapeData *tdp = &tapeDatas;
	assert((tdp->state & (TS_TAPIN | TS_TZXIN | TS_VOCIN)) != 0 && (tdp->state & (TS_PLAY | TS_ERRIN | TS_ENDIN)) == 0);
	tdp->state |= TS_PLAY;
	TapeInGetBit = TzxInGetBit;
	TzxInPlay();
	tdp->in.event = EmulAddEvent((EmulEventFunc)tapeInGetBit, (void *)*TapeInGetBit, TAPE_TIMER_VALUE);
	VtpUpdate();
	if ((tdp->state & TS_FASTLOAD) != 0)
	{
		CFGSET(fastmode, CFGGET(fastmode) | 2);
#ifdef REAL_AUDIO
		AudioControl(AUC_OFF);
#endif
	}
}
void
TapeInStop(void)
{
	tapeData *tdp = &tapeDatas;
	assert((tdp->state & TS_PLAY) != 0);
	tdp->state &= ~TS_PLAY;
	TapeInGetBit = tapeInGetBitDummy;
	EmulRemoveEvent(tdp->in.event, 0);
	VtpUpdate();
	if ((tdp->state & TS_FASTLOAD) != 0)
	{
		CFGSET(fastmode, CFGGET(fastmode) & ~2);
#ifdef REAL_AUDIO
		if (CFGGET(fastmode) == 0)
		{
			AudioControl(AUC_ON);
		}
#endif
	}
}
static void
tapeInGetBit(uns8 (*func)(void))
{
	(void)func();
	if ((tapeDatas.state & TS_PLAY) != 0)
	{
		tapeDatas.in.event = EmulAddEvent((EmulEventFunc)tapeInGetBit, (void *)func, TAPE_TIMER_VALUE);
	}
}
static uns8
tapeInGetBitDummy(void)
{
	return 0;
}
const char *
TapeOutGetName(void)
{
	return tapeDatas.out.fname;
}
#ifndef NO_ED_TRAPS
void
TapeOutWriteBlock(void)
{
	tapeData *tdp = &tapeDatas;
	DEB(Msg(M_DEBUG, "%stape saving: addr = %04X, length = %04X", A ? "headerless " : "", IX, DE););
	if ((tdp->state & TS_TAPOUT) == 0)
	{
		VtpNewOutputFile();
	}
	if ((tdp->state & TS_TAPOUT) != 0)
	{
		TapOutWriteBlock(tdp->out.fp);
		return;
	}
	TSTATES += 10 - 8;
	DEC_R();
	HL = 0x1f80;
	PC++;
}
#endif
int
TapeOutAttach(char *fname)
{
	tapeData *tdp = &tapeDatas;
	int ret = -1;
	TapeOutDetach();
	if ((tdp->out.fp = Fopen(fname, "wb")) == (FILE *)0)
	{
		Msg(M_PERR, "couldn't open <%s> for writing", fname);
	}
	else
	{
		tdp->state |= TS_TAPOUT;
		tdp->out.fname = fname;
		ret = 0;
	}
	return ret;
}
void
TapeOutDetach(void)
{
	tapeData *tdp = &tapeDatas;
	if ((tdp->state & TS_TAPOUT) != 0)
	{
		tdp->state &= ~(TS_TAPOUT | TS_ERROUT | TS_ENDOUT);
		(void)fclose(tdp->out.fp);
		free((char *)(tdp->out.fname));
		tdp->out.fname = (char *)0;
	}
}
static void
tapeCatchFault(int signo)
{
	longjmp(tapeEnv, -1);
}
